import numpy as np
from numpy.lib.stride_tricks import as_strided as ast

def norm_shape(shape):
    '''
    Normalize numpy array shapes so they're always expressed as a tuple, 
    even for one-dimensional shapes.
    
    Parameters
        shape - an int, or a tuple of ints
    
    Returns
        a shape tuple
    '''
    try:
        i = int(shape)
        return (i,)
    except TypeError:
        # shape was not a number
        pass

    try:
        t = tuple(shape)
        return t
    except TypeError:
        # shape was not iterable
        pass
    
    raise TypeError('shape must be an int, or a tuple of ints')


def sliding_window_1d(a,ws,ss = None):
    '''
    Parameters
        a  - a 1D array
        ws - the window size, in samples
        ss - the step size, in samples. If not provided, window and step size
             are equal. 
    '''
    
    if None is ss:
        # no step size was provided. Return non-overlapping windows
        ss = ws
    
    # calculate the number of windows to return, ignoring leftover samples, and
    # allocate memory to contain the samples
    valid = len(a) - ws
    nw = (valid) // ss
    out = np.ndarray((nw,ws),dtype = a.dtype)
    
    for i in xrange(nw):
        # "slide" the window along the samples
        start = i * ss
        stop = start + ws
        out[i] = a[start : stop]
    
    return out

from itertools import product
def sliding_window_nd(a,ws,ss = None):
    '''
    Return a sliding window over a in any number of dimensions
    
    Parameters:
        a  - an n-dimensional numpy array
        ws - an int (a is 1D) or tuple (a is 2D or greater) representing the size 
             of each dimension of the window
        ss - an int (a is 1D) or tuple (a is 2D or greater) representing the 
             amount to slide the window in each dimension. If not specified, it
             defaults to ws.
    
    Returns
        an array containing each n-dimensional window from a
    '''
    if None is ss:
        # ss was not provided. the windows will not overlap in any direction.
        ss = ws
    
    # ensure that window and step sizes are expressed as tuples, even if they're
    # one-dimensional
    ws = norm_shape(ws)
    ss = norm_shape(ss)
    
    # convert ws, ss, and a.shape to numpy arrays so that we can do math in every 
    # dimension at once.
    ws = np.array(ws)
    ss = np.array(ss)
    shape = np.array(a.shape)
    
    
    # ensure that ws, ss, and a.shape all have the same number of dimensions
    ls = [len(shape),len(ws),len(ss)]
    if 1 != len(set(ls)):
        raise ValueError(\
        'a.shape, ws and ss must all have the same length. They were %s' % str(ls))
    
    # ensure that ws is smaller than a in every dimension
    if np.any(ws > shape):
        raise ValueError(\
        'ws cannot be larger than a in any dimension.\
 a.shape was %s and ws was %s' % (str(a.shape),str(ws)))
    
    # For each dimension, create a list of all valid slices
    slices = [[] for i in range(len(ws))]
    for i in xrange(len(ws)):
        nslices = ((shape[i] - ws[i]) // ss[i]) + 1
        for j in xrange(0,nslices):
            start = j * ss[i]
            stop = start + ws[i]
            slices[i].append(slice(start,stop))
    # Get an iterator over all valid n-dimensional slices of the input
    allslices = product(*slices)
    
    # Allocate memory to hold all valid n-dimensional slices
    nslices = np.product([len(s) for s in slices])
    out = np.ndarray((nslices,) + tuple(ws),dtype = a.dtype)
    for i,s in enumerate(allslices):
        out[i] = a[s]
    
    
    return out

def sliding_window(a,ws,ss = None,flatten = True):
    '''
    Return a sliding window over a in any number of dimensions
    
    Parameters:
        a  - an n-dimensional numpy array
        ws - an int (a is 1D) or tuple (a is 2D or greater) representing the size 
             of each dimension of the window
        ss - an int (a is 1D) or tuple (a is 2D or greater) representing the 
             amount to slide the window in each dimension. If not specified, it
             defaults to ws.
        flatten - if True, all slices are flattened, otherwise, there is an 
                  extra dimension for each dimension of the input.
    
    Returns
        an array containing each n-dimensional window from a
    '''
    
    if None is ss:
        # ss was not provided. the windows will not overlap in any direction.
        ss = ws
    ws = norm_shape(ws)
    ss = norm_shape(ss)
    
    # convert ws, ss, and a.shape to numpy arrays so that we can do math in every 
    # dimension at once.
    ws = np.array(ws)
    ss = np.array(ss)
    shape = np.array(a.shape)
    
    
    # ensure that ws, ss, and a.shape all have the same number of dimensions
    ls = [len(shape),len(ws),len(ss)]
    if 1 != len(set(ls)):
        raise ValueError(\
        'a.shape, ws and ss must all have the same length. They were %s' % str(ls))
    
    # ensure that ws is smaller than a in every dimension
    if np.any(ws > shape):
        raise ValueError(\
        'ws cannot be larger than a in any dimension.\
 a.shape was %s and ws was %s' % (str(a.shape),str(ws)))
    
    # how many slices will there be in each dimension?
    newshape = norm_shape(((shape - ws) // ss) + 1)
    # the shape of the strided array will be the number of slices in each dimension
    # plus the shape of the window (tuple addition)
    newshape += norm_shape(ws)
    # the strides tuple will be the array's strides multiplied by step size, plus
    # the array's strides (tuple addition)
    newstrides = norm_shape(np.array(a.strides) * ss) + a.strides
    strided = ast(a,shape = newshape,strides = newstrides)
    if not flatten:
        return strided
    
    # Collapse strided so that it has one more dimension than the window.  I.e.,
    # the new array is a flat list of slices.
    meat = len(ws) if ws.shape else 0
    firstdim = (np.product(newshape[:-meat]),) if ws.shape else ()
    dim = firstdim + (newshape[-meat:])
    # remove any dimensions with size 1
    dim = filter(lambda i : i != 1,dim)
    return strided.reshape(dim)
    


import numpy as np
from timeit import timeit

## Let's use a sliding window that overlaps by half
#statement = 'sliding_window(a,2048,1024)'
## Slide the window over 5 seconds of "audio" at 44.1Khz sampling rate
#setup = 'from __main__ import sliding_window,np; a = np.zeros(44100 * 5)'
#print timeit(statement,setup,number = 1000)
#statement = 'sliding_window(a,(10,9),(5,4))'
#setup = 'from __main__ import sliding_window,np; a = np.zeros((1000,1000))'
#print timeit(statement,setup,number = 1000)



ws = 10
step = 5
nd_seconds = []
stride_seconds = []
ratio = []
while ws < 500:
    shape = (1000,1000)
    window = (ws,) * 2
    ss = (ws / 2,) * 2

    
    statement = 'sliding_window_nd(a,%s,%s)' % (str(window),str(ss))
    setup = 'from __main__ import sliding_window_nd,np; a = np.zeros(%s)' % str(shape)
    nd_seconds.append(timeit(statement,setup,number = 100))
    
    statement = 'sliding_window(a,%s,%s)' % (str(window),str(ss))
    setup = 'from __main__ import sliding_window,np; a = np.zeros(%s)' % str(shape)
    stride_seconds.append(timeit(statement,setup,number = 100))
    ratio.append(ws / 1000.)
    print ratio[-1]
    ws += step


#ws = 100
#step = 50
#
#nd_seconds = []
#stride_seconds = []
#ratio = []
#while ws < 5000:
#    shape = 10000
#    window = (ws,)
#    ss = (ws / 2,)
#    statement = 'sliding_window_nd(a,%s,%s)' % (str(window),str(ss))
#    setup = 'from __main__ import sliding_window_nd,np; a = np.zeros(%s)' % str(shape)
#    nd_seconds.append(timeit(statement,setup,number = 1000))
#    
#    statement = 'sliding_window(a,%s,%s)' % (str(window),str(ss))
#    setup = 'from __main__ import sliding_window,np; a = np.zeros(%s)' % str(shape)
#    stride_seconds.append(timeit(statement,setup,number = 1000))
#    ratio.append(ws / float(shape))
#    print ratio[-1]
#    ws += step

print len(ratio)

from matplotlib import pyplot as plt

a = np.array([nd_seconds,stride_seconds]).T
p1 = plt.plot(nd_seconds)
p2 = plt.plot(stride_seconds)
plt.legend([p1,p2],['for-loop','stride tricks'])
plt.ylabel('seconds')
plt.xlabel('array-to-window size ratio')
plt.xticks(np.arange(0,100,20),ratio[::20])
plt.show()
plt.savefig('2D.png')
plt.clf()

