from __future__ import division
import os.path
from random import choice
from time import time

import numpy as np

from config import *
from scikits.audiolab import Sndfile,play
from zounds.model.framesearch import *
from zounds.environment import Environment
from zounds.acquire.acquirer import audio_files
from zounds.nputil import pad
from zounds.analyze.audiostream import read_frames_mono
from zounds.analyze.resample import Resample
from zounds.visualize.plot import plot
from zounds.util import SearchSetup


class Zndfile(Sndfile):
    '''
    A handy wrapper around scikits.audiolab.Sndfile which allows clients to 
    address sections of the file in a samplerate-agnostic way. Additionally,
    it performs on-the-fly resampling to the Zounds environment's sample rate.
    '''
    def __init__(self,filepath,out_samplerate = Z.samplerate):
        '''
        out_samplerate - the sample rate that all samples read by this instance
                         should be returned in, if resampling is requested when
                         reading
        '''
        Sndfile.__init__(self,filepath)
        self._out_samplerate = out_samplerate
    
    @property
    def nseconds(self):
        '''
        The length of this audio file, in seconds
        '''
        return self.nframes / self.samplerate
    
    @property
    def need_resample(self):
        '''
        Return true if samples read should be resampled
        '''
        return self.samplerate != self._out_samplerate
    
    def read_frames_padded(self,total_length_secs,resample = True):
        '''
        Read frames from the file, ensuring that the segment returned is at
        least total_length_secs in length
        ''' 
        # total desired length in samples, at the file's sampling rate
        ts = total_length_secs * self.samplerate
        # total number of frames to read from the file
        nframes = self.nframes if total_length_secs > self.nseconds else ts
        # Read samples.  Sum to mono if necessary, and pad with zeros so the 
        # length of samples is ts
        samples = pad(read_frames_mono(self,nframes),ts)
        if resample and self.need_resample:
            # resampling was requested, and the output sample rate doesn't match
            # the file's sample rate.
            return \
            Resample(self.samplerate,self._out_samplerate).all_at_once(samples)
        
        # Resampling wasn't requested, or wasn't necessary. Return the samples
        # unmodified
        return samples
    
    def seek_seconds(self,seconds,whence = 0):
        '''
        Move n seconds from either 0 = beginning of file, 1 = current position,
        or 2 = end of file
        '''
        nsamples = seconds * self.samplerate
        self.seek(nsamples,whence) 

    

def get_query(path,files,minseconds,maxseconds):
    # choose a random audio file from path
    snd = Zndfile(os.path.join(path,choice(files)))
    if snd.nseconds < minseconds:
        # The lenght of the chosen sound is less than minseconds.  Return the 
        #  whole sound, padded with zeros so it's minseconds long, and sampled
        # at the current environment's sampling rate.
        return snd.read_frames_padded(minseconds)
    
    if snd.nseconds > minseconds and snd.nseconds < maxseconds:
        maxseconds = snd.nseconds
    
    # choose a length in seconds for the query that falls between minseconds
    # and maxseconds
    lsecs = minseconds + (np.random.random() * (maxseconds - minseconds))
    # choose the starting point for the segment
    start = np.random.random() * (snd.nseconds - lsecs)
    # fetch the segment at the current environment's sampling rate
    snd.seek_seconds(start)
    return snd.read_frames_padded(lsecs)
    
    
if __name__ == '__main__':
    
    args,search = SearchSetup(FrameModel).setup()
    
    # Pick a sound segment
    d = args.sounddir
    files = audio_files(d)
    
    while True:
        # get a random query
        query = get_query(d,files,args.minseconds,args.maxseconds)
        # Do the search and time it
        starttime = time()
        queryframes,results = search.search(query,args.nresults)
        print 'search took %1.4f seconds' % (time() - starttime)
        # play the query sound
        Z.synth.playraw(query)
        # play the search results
        for _id,addr in results:
            try:
                print _id,addr
                Environment.instance.play(FrameModel[addr].audio)
            except KeyboardInterrupt:
                continue
        
        # TODO: Visualize the distance between
        if raw_input('visualize_rersults? (y or n)') == 'y':
            plot(queryframes[args.feature],'query')
            to_delete = ['query.png']
            for i,r in enumerate(results):
                _id,addr = r
                frames = FrameModel[addr]
                fn = '%s_%i' % (frames.external_id[0],i)
                plot(frames[args.feature],fn)
                to_delete.append('%s.png' % fn)
            raw_input('press any key to continue')
            for delete in to_delete:
                os.remove(delete)